# main.py
# ToDo Manager
# This software is released under the Python 2.0.1 license
#
# Author: Brian Bernas
#
# The core application calls

import locale
import os, sys, time

# Set the locale to the user's setting
try:
	locale.setlocale(locale.LC_ALL, '')
	lang = locale.getdefaultlocale()[0]

	if not os.environ.has_key('LANG') and lang:
		os.environ['LANG'] = lang

except (locale.Error, AttributeError, ValueError):
	pass

from __init__ import *
from ConfigParser import ConfigParser, NoSectionError
from tdmcalls import *
import plugin

import thread, traceback

# Add the main source directory to the path for import operations
sys.path.append(get_root_directory())

# gettext setup
_ = get_translation('core')

try:
	import webbrowser
except ImportError:
	import BrowserControl
	webbrowser = BrowserControl
	del BrowserControl

TRUE = 1
FALSE = 0

ALL = "ALL"

# Config variables
_new_file = FALSE
_load_file = None
_load_config = TRUE
_load_plugins = TRUE

help_file = os.path.join(get_root_directory(), 'docs', 'index.html')

#-------------------------------------------------------------------------------
class App:
	"""The core ToDo Manager class. All of the major functions are included here."""
	def __init__(self):
		# Variables
		self.__interfaces = []	# Multiple interface bindings, I've gone mad
		self.__tasks = []
		self.__current_file = None
		self.__file_modified = FALSE
		self.__options = {}
		self.__config_file = None
		self.__default_file = None
		self.__tmpfile = None

		# See if the program should continue executing after the args have been read
		if not parse_args(sys.argv[1:]):
			sys.exit(0)

		# Get the config file and the default file
		wkdir = None
		userdir = get_user_directory()

		# Verify that the directory exists
		if os.path.isdir(userdir):
			wkdir = userdir
		else:
			try:
				os.mkdir(userdir)
			except OSError:
				# No write permission to userdir
				userdir = None

		if not wkdir:
			wkdir = get_root_directory()

		self.__config_file = os.path.join(wkdir, 'todo-manager.ini')
		self.__default_file = os.path.join(wkdir, 'default.tmf')

		# init stuff goes here (load prefs blah, blah...)
		if _load_config:
			self.__load_config()

		if self.get_setting("Core", "loadplugins", TRUE) and _load_plugins:
			# Get the plugins in the root directory
			plgs = [os.path.join(get_root_directory(), 'plugins')]

			# Get plugins from the user's directory
			if get_user_directory() != get_root_directory():
				plgs.append(os.path.join(get_user_directory(), 'plugins'))

			plugin.fmttime = self.fmttime # This gives the plugins access to fmttime
			plugin.load_all_plugins(plgs)

		if _load_file:
			self.__ext_load_file(_load_file)
		elif _new_file:
			self.new_file()
		else:
			self.load_default_file()

		# Start the thread that checks for due tasks
		thread.start_new_thread(self.__check_due_tasks, ())

		if userdir and (userdir != wkdir):
			self.__config_file = os.path.join(userdir, 'todo-manager.ini')
			self.__current_file = self.__default_file = os.path.join(userdir, 'default.tmf')
			self.__file_modified = TRUE

		# Keep track if a temp file was found
		self.__found_tmpfile = self.__locate_tmpfile()

	def __create_new_task(self, name=""):
		task = {
			"Name": name,
			"Notes": "",
			"Finished": FALSE,
			"DisableFinish": FALSE,
			"FinishedDate": "",
			"DueDate": "",
			"Priority": 1,
			"Group": [],
			}

		return task

	def add_task(self, name):
		"""Add a task to the list. It takes one argument, the name of the task."""

		name = unicode_to_utf8(name)

		# Does the task already exist? The world may never know...
		for t in self.__tasks:
			if t["Name"] == name:
				return FALSE

		task = self.__create_new_task(name)

		self.__tasks.append(task)

		self.__save_tmpfile()

		self.__file_modified = TRUE

		# Tell the interfaces that there is a new item
		self.interface_call(ALL, "insert_list_item", name)

		return TRUE

	# The interfaces should call this. They should never has plain access to the tasks.
	def set_task_value(self, name, key, value, ui=None):
		"""Set a value of a task. The first argument is the task name, second
		is the setting key, and the last one is the value that the key is set to."""

		name = unicode_to_utf8(name)
		value = unicode_to_utf8(value) 

		timefmt = self.get_setting("Core", "timeformat", '')

		found = FALSE
		for t in self.__tasks:
			if t["Name"] == name:
				found = TRUE
				break

		# Error checking if the task doesn't exist
		if not found: return FALSE

		if t.has_key(key) and value != None:
			if key == "Priority":
				value = int(value)
				if value < 1: value = 1
				elif value > 5: value = 5
			elif key == "Finished":
				if not t["DisableFinish"]:
					value = int(value)
					ft = ("", curtime())[value]
					t["FinishedDate"] = ft
				else: value = FALSE
			elif key == "DisableFinish":
				value = int(value)
				if value == TRUE:
					t["DueDate"] = ""
					t["Finished"] = FALSE
					t["FinishedDate"] = ""
			elif key == "DueDate" and value:
				# Check the format of the due date that the user has entered
				value = check_time_values(strptime(timefmt, value))
				if not value:
					# The format is invalid
					self.interface_call(ui, "display_error", _("Invalid Format"), _("The due date "\
						"is not in the correct format.\nThe format is %s") %self.fmttime(curtime()))
					return FALSE
			elif key == "Group":
				# Split multiple groups into a list
				x = value.split(',')

				# Eliminate unwanted white space
				# Tasks can't be in the group 'ALL', this is a global group
				for i in range(len(x)):
					x[i] = x[i].strip()

				# Remove invalid groups
				if 'ALL' in x: x.remove('ALL')
				if '' in x: x.remove('')

				x.sort()
				value = x

			t[key] = value

			# So the interface is sent a formatted time
			if key == "DueDate":
				value = self.fmttime(value)

			self.interface_call(ALL, "set_list_item_value", name, key.lower(), value)

			# If the changed value effects the finished date
			if key in ("Finished", "DisableFinish"):
				self.interface_call(ALL, "set_list_item_value", name, "finisheddate",
					self.fmttime(t["FinishedDate"]))

			if key == "Group":
				self.interface_call(ALL, "update_list_groups", self.__get_task_groups())

			self.__save_tmpfile()

			self.__file_modified = TRUE
			return TRUE

		return FALSE

	def get_task_value(self, name, key):
		"""Get a value from the task. The first argument is the task name, the
		second is the setting key."""

		for t in self.__tasks:
			if t["Name"] == unicode_to_utf8(name):
				if t.has_key(key):
					v = t[key]

					# Format the dates
					if (key == "DueDate" or key == "FinishedDate"):
						if v:
							v = self.fmttime(v)
						else: v = ''
					# Build group lists
					if key == "Group":
						x = ''
						gsize = len(t["Group"])
						for i in range(gsize):
							x = "%s%s" %(x, t["Group"][i])
							if i < gsize-1: x = x + ','
							else: break
						v = x

					return v

	def __get_task_names(self):
		"""Return a tuple of the task names"""

		names = []
		for t in self.__tasks:
			names.append(t["Name"])

		return names

	def delete_task(self, name):
		"""Remove a task from the list. The only argument is the task index."""

		name = unicode_to_utf8(name)

		for t in self.__tasks:
			if t["Name"] == name:
				self.__tasks.remove(t)

				self.__save_tmpfile()

				self.__file_modified = TRUE

				self.interface_call(ALL, "remove_list_item", name)
				self.interface_call(ALL, "update_list_groups", self.__get_task_groups())


				return TRUE

		return FALSE

	def new_file(self, ui=None):
		"""Unload any file that is loaded and start fresh."""

		if self.ask_save_curfile(ui) == -1:
			return

		# Clear out all the current tasks
		# This is better than just removing each task from self.__tasks
		del self.__tasks
		self.__tasks = []

		self.__current_file = None
		self.__file_modified = FALSE

		self.__clear_tmpfile()

		self.interface_call(ALL, "clear_list_items")
		self.interface_call(ALL, "update_list_groups", [])
		self.interface_call(ALL, "set_current_filename", _("Untitled"))

	def do_file_save(self, filename=None, ui=None):
		if not filename:
			filename = self.__current_file

		# This will cause a saveas prompt to be displayed if the current file is the default file.
		if filename == self.__default_file:
			filename = None

		return self.__ext_save_file(filename, ui=ui)

	def do_file_saveas(self, ui):
		return self.__ext_save_file(ui=ui)

	def __ext_save_file(self, filename=None, func=None, desc=None, ext=None, tasks=None, ui=None):
		if not tasks: tasks = self.__dup_task_list()

		if not func:
			desc = _("ToDo Manager File")
			ext = "*.tmf"

		if not filename:
			filename = self.interface_call(ui, "file_saveas_prompt", desc, ext)
			if not filename: return FALSE

		if not os.path.splitext(filename)[1]:
			filename = "%s%s" %(filename, ext[1:])

		try:
			file = open(filename, 'wb')
		except IOError:
			self.interface_call(ui, "display_error", _("File Error"),
				_("The file can't be written. You may not be permitted to write to the drive."))
			return FALSE

		if not func:
			func = self.__save_file
			self.__current_file = filename
			self.__file_modified = FALSE
			self.interface_call(ALL, "set_current_filename", os.path.split(filename)[1])

		# Execute the function
		func(tasks, file)

		file.close()

		return TRUE

	def __save_file(self, tasks, file):
		"""Save the current task list to a file."""

		# Format the due date and the finished date
		for t in tasks:
			t["DueDate"] = storetime(t["DueDate"])
			t["FinishedDate"] = storetime(t["FinishedDate"])

			# Convert the groups to a string separated by commas
			x = ''
			gsize = len(t["Group"])
			for i in range(gsize):
				x = "%s%s" %(x, t["Group"][i])
				if i < gsize-1: x = x + ','
				else: break
			t["Group"] = x

		# Create the TaskFile
		tf = TaskFile(file)

		# Save each task
		for t in tasks:
			tf.save_task(t)

	def __save_tmpfile(self):
		if self.get_setting("Core", "savebackup", FALSE):
			if not self.__tmpfile:
				# Generate a temp filename
				dat = os.path.join(get_user_directory(), "tmp%d" %int(time.time()))
				self.__tmpfile = "%s.tmf" %dat

				# Write the data file. This stores extra info about the backup
				file = open(dat, 'w')
				# Write the filename
				curfile = ("None\n", "%s\n" %self.__current_file)[self.__current_file != None]
				file.write(curfile)
				file.close()

			# Write the file
			file = open(self.__tmpfile, 'wb')
			self.__save_file(self.__dup_task_list(), file)
			file.close()

	def __clear_tmpfile(self):
		if self.__tmpfile:
			if os.path.isfile(self.__tmpfile):
				# Remove the file
				os.remove(self.__tmpfile)
				# Remove the data file
				os.remove(os.path.splitext(self.__tmpfile)[0])
				self.__tmpfile = None

	def __locate_tmpfile(self):
		# The list that is returned after the scan
		# Name of file, Name of backup, recently modified
		retlist = []

		# Get a list of all files in the directory
		l = os.listdir(get_user_directory())

		# This should ask the user if they want to restore all tmp files found
		for f in l:
			# Find the temp data file
			if f[:3] == 'tmp' and not os.path.splitext(f)[1]:
				# Get the name of the file
				tmpfile = os.path.join(get_user_directory(), f)
				file = open(tmpfile, 'r')
				# Read the source filename from this file
				origfile = file.read().strip()
				file.close()

				# Make sure the files still exists
				tmpfile = tmpfile + '.tmf' # add the extension
				if os.path.isfile(origfile) and os.path.isfile(tmpfile):
					# Compare modification times
					# If the original has been modified then don't prompt for a restore
					ofm = os.path.getmtime(origfile)
					bfm = os.path.getmtime(tmpfile)

					# Add the items to the list
					retlist.append([origfile, tmpfile, bfm > ofm])

		return retlist

	def __prompt_tmpfile(self, ui):
		for f in self.__found_tmpfile:
			i = 0
			if f[2]:
				i = self.interface_call(ui, "ask_yes_no", _("Restore Backup File"),
					_("A backup of the file %s was found. Would you like to recover all "\
					"changes for this file?") %f[0])

				# If the user responded "yes" then restore the file
				if i: self.__restore_tmpfile(f[1], f[0])

			else:
				i = not self.interface_call(ui, "ask_yes_no", _("Backup File Warning"),
					_("A backup file for %s was found at %s but cannot be restored because "\
					"the original file has been modified since the backup file was created. "\
					"would you like to preserve the backup file anyway?") %(f[0], f[1]))

			# Delete the tmp files
			if i:
				os.remove(os.path.splitext(f[1])[0])
				os.remove(f[1])

		# Set this back to a blank list so it isn't given to other ui's
		self.__found_tmpfile = []

	def __restore_tmpfile(self, tmpfile, origfile):
		# Read the backup
		file = open(tmpfile, 'r')
		self.__load_file(file)
		file.close()

		# Write to the original
		file = open(origfile, 'wb')
		self.__save_file(self.__dup_task_list(), file)
		file.close()

		self.__current_file = origfile

	def do_file_open(self, filename=None, ui=None):
		self.__ext_load_file(filename=filename, ui=ui)

	def __ext_load_file(self, filename=None, func=None, desc=None, ext=None, ui=None):
		"""Extended file loading function."""

		success = FALSE	# If the file was successfully loaded

		if not func:
			desc = _("ToDo Manager File")
			ext = "*.tmf"

		if not filename:
			filename = self.interface_call(ui, "file_open_prompt", desc, ext)
			if not filename: return

		if self.ask_save_curfile(ui) == -1:
			return

		# Make sure the file exists before opening it
		if not os.path.isfile(filename):
			self.interface_call(ui, "display_error", _("File Error"),
				_("The file %s does not exist.") %filename)
			return

		file = open(filename, 'r')

		if not func:
			self.__load_file(file)
			self.__current_file = filename
			success = TRUE
		else:
			# if the user tries to load the wrong filetype, this will throw and exception
			try:
				tasks = func(file)
				del self.__tasks
				self.__tasks = tasks
				success = TRUE
			except:
				# Print a traceback
				traceback.print_exc()
				self.interface_call(ui, "display_error", _("Invalid Format"),
					_("The file %s is in an invalid format or "\
					"is not readable by this plugin.") %filename)

		file.close()

		# The rest of the function only needs to be run if the file was loaded
		if not success: return

		self.__file_modified = FALSE

		self.__clear_tmpfile()

		self.interface_call(ALL, "update_list_items", self.__get_task_names())
		self.interface_call(ALL, "update_list_groups", self.__get_task_groups())

		# See if the file that has just been loaded is the default file
		f = os.path.join(get_root_directory(), filename)
		if f == self.__default_file: val = 'DEFAULT'
		else: val = os.path.split(filename)[1]

		self.interface_call(ALL, "set_current_filename", val)

	def __load_file(self, file):
		"""Load a task list from a file. Only argument is the file."""

		del self.__tasks
		self.__tasks = []

		tf = TaskFile(file)
		while 1:
			attrs = tf.read_next_task()
			if not attrs: break
			new_task = self.__create_new_task()
			for a, v in attrs:
				# Some values should be integers
				if a in ("Priority", "Finished", "DisableFinish"):
					v = int(v)
				elif a in ("DueDate", "FinishedDate"):
					v = parse_time(v)
				elif a == "Group":
					v = unicode_to_utf8(v)

					if v:
						x = v.split(',')

						for i in range(len(x)):
							x[i] = x[i].strip()

						if '' in x: x.remove('')

						v = x
					else: v = []

				new_task[a] = unicode_to_utf8(v)

			self.__tasks.append(new_task)

	def load_default_file(self, ui=None):
		"""Load the default file."""

		if os.path.isfile(self.__default_file):
			self.__ext_load_file(self.__default_file, ui=ui)
		else:
			# The file doesn't exist but still try to write to it
			self.__current_file = self.__default_file

	def __load_config(self):
		"""Load the configuration file. This should not be called by anything but
		this class."""

		global _new_file, _load_file
		cp = ConfigParser()
		cp.read(self.__config_file)

		options = self.__options
		for sec in cp.sections():
			options[sec] = {}
			for opt in cp.options(sec):
				val = cp.get(sec, opt)
				try:
					# Attempt to convert the value to an integer
					val = int(val)
				except ValueError:
					pass

				if val == "None": val = None

				options[sec][opt] = val

		# The options specific to the core
		val = self.get_setting("Core", "startupfile", 0)
		if val == 1 and not _load_file:
			file = self.get_setting("Core", "lastfileloaded", None)
			if file != "None":
				_load_file = file
		elif val == 2:
			_new_file = TRUE

		val = self.get_setting("Core", "timeformat", None)
		# Attempt to get the locale-specific time format
		if not val:
			try:
				format = locale.nl_langinfo(locale.D_FMT)
				format = format + ' ' + locale.nl_langinfo(locale.T_FMT)
			except:
				format = "%Y-%m-%d %H:%M:%S"

			self.save_setting("Core", "timeformat", format)

	def __save_config(self):
		"""Save the configuration file to disk."""

		# Core options
		if self.get_setting("Core", "startupfile", 0) == 1:
			if self.__current_file != self.__default_file:
				self.save_setting("Core", "lastfileloaded", self.__current_file)

		options = self.__options
		cp = ConfigParser()

		for sec in options.keys():
			cp.add_section(sec)
			for opt in options[sec].keys():
				cp.set(sec, opt, options[sec][opt])

		try:
			file = open(self.__config_file, 'w')
			cp.write(file)
			file.close()
		except IOError:
			pass

	def save_setting(self, section, option, value):
		"""Save a setting to the setting handler. The first argument is the
		section, second is the option name, and last is the value."""

		if not self.__options.has_key(section):
			self.__options[section] = {}

		self.__options[section][option] = value

	def get_setting(self, section, option, default=None):
		"""Retrieve a setting. The first arguemnt is the section, second is the
		option name, and last is the default value is the setting is not found."""

		if not self.__options.has_key(section):
			return default
		if not self.__options[section].has_key(option):
			return default
		return self.__options[section][option]

	def bind_interface(self, interface):
		"""Bind a interface to the call system."""

		if not interface in self.__interfaces:
			self.__interfaces.append(interface)

			# If a temp file was found then tell the first interface
			self.__prompt_tmpfile(interface)

			# Tell the interface a few things it should know

			# Current filename
			file = self.__current_file
			if file:
				if file == self.__default_file: file = "DEFAULT"
				else: file = os.path.split(file)[1]
			else: file = _("Untitled")

			self.interface_call(interface, "set_current_filename", file)

			# The contents of the task list
			self.interface_call(interface, "update_list_items", self.__get_task_names())
			self.interface_call(interface, "update_list_groups", self.__get_task_groups())

			# Plugins that are loaded
			plgs = self.get_plugin_info()

			for p in plgs:
				self.interface_call(interface, "add_plugin", p)

			# The interface was bound successfully 
			return TRUE

		return FALSE

	def unbind_interface(self, interface):
		"""Unbind an interface from the core."""

		if interface in self.__interfaces:
			# If this is the only interface then ask if the current file should be saved
			if len(self.__interfaces) == 1:
				# If cancel was selected then stop
				if self.ask_save_curfile(interface) == -1: return FALSE

			self.interface_call(interface, "save_settings")
			self.interface_call(interface, "clear_list_items")
			self.interface_call(interface, "update_list_groups", [])
			self.__interfaces.remove(interface)
		
		# If there aren't any interfaces then call quit
		if not self.__interfaces:
			self.__quit()

		return TRUE

	def show_html_help(self, ui=None):
		"""Display the HTML help file."""

		self.__open_url(help_file, ui)

	def show_home_page(self, ui=None):
		self.__open_url("http://todo-manager.sourceforge.net", ui)
		
	def __open_url(self, url, ui=None):
		# If the user specified a web browser then use it if possible
		browser = self.get_setting("Core", "webbrowser", None)
		if browser:
			try:
				thread.start_new_thread(os.system, (browser %url,))
			except TypeError:
				self.interface_call(ui, "display_error", _("Error"),
					_("The Web Browser command is not in the right format.\n"\
					"It should be in the format: command \"%s\".\nPlease correct this error."))
			return

		try:
			webbrowser.open(url)
		except Exception:
			self.interface_call(ui, "display_error", _("Browser Error"),
				_("A runnable web browser could not be found."))

	def ask_save_curfile(self, ui=None):
		"""Prompt the user to save the current file."""

		if self.__file_modified:
			# If it's the default file save it and move on
			if self.__current_file == self.__default_file:
				self.__ext_save_file(self.__default_file, ui=ui)
				return

			x = self.__current_file
			if not x: x = _("Untitled")

			ans = self.interface_call(ui, "ask_yes_no_cancel", _("Current File"),
				_("Do you want to save file %s?") %x)

			if ans == -1: return -1	# The user has selected cancel
			elif ans == 0: return 0	# Do not save
			elif ans == 1:	# Save
				if not self.do_file_save(ui=ui): return -1	# Cancel was clicked on the file dialog
				return 1

	def is_task_due(self, name):
		"""Checks to see if the task is due. The only argument is the task index."""

		# Returns a tuple containing two integers: first is if date is due, second is if time is due
		# Return: 0=due; 1=overdue; -1=not due

		found = FALSE
		for task in self.__tasks:
			if task["Name"] == name:
				found = TRUE
				break

		# If the task wasn't found, then cut the function short
		if not found: return (-1, -1)

		dd = task["DueDate"]
		if not dd or task["Finished"]: return (-1, -1)

		# Doesn't include time info
		dd2 = check_time_values(dd[:3])
		val1 = cmptime(dd2, check_time_values(curtime()[:3]))

		if dd[3] != 0 or dd[4] != 0 or dd[5] != 0:
			val2 = cmpcurtime(dd)
		else:
			val2 = -1

		return (val1, val2)

	def __check_due_tasks(self):
		"""Check all of the tasks to see if they are due."""

		while 1:
			time.sleep(60.0)
			for n in self.__get_task_names():
				# If the task is due, tell the interface
				if self.is_task_due(n)[1] == 0:
					self.interface_call(ALL, "display_warning",_("A Task Is Due"),
						_("The task \"%s\" is due.") %n)

	def exec_plugin(self, index, ui=None, *args, **kw):
		"""Execute a plugin"""

		plgs = plugin._loaded_plugins

		found = FALSE
		for p in plgs:
			if p['value'] == index:
				found = TRUE
				break

		# If the plugin wasn't found then just return
		if not found: return

		if kw.has_key("tasks"):
			# Get the tasks that the user wants to use
			new_tasks = []
			for n in kw["tasks"]:
				for t in self.__tasks:
					if t["Name"] == n:
						new_tasks.append(t.copy())
		else:
			new_tasks = self.__dup_task_list()

		if p['type'] == 'file_import':
			self.__ext_load_file(func=p['function'], desc=p['name'], ext=p['format'], ui=ui)

		elif p['type'] == 'file_export':
			# Convert the time structures into strings
			for i in range(len(new_tasks)):
				dd = new_tasks[i]["DueDate"]
				fd = new_tasks[i]["FinishedDate"]

				if dd != '':
					new_tasks[i]["DueDate"] = self.fmttime(dd)
				if fd != '':
					new_tasks[i]["FinishedDate"] = self.fmttime(fd)
				
			self.__ext_save_file(func=p['function'], desc=p['name'], ext=p['format'],
				tasks=new_tasks, ui=ui)

		elif (p['type'] == 'task_sorting') or (p['type'] == 'task_filter'):
			return p['function'](self.__dup_task_list())

	def get_plugin_info(self):
		"""Return information on all available plugins"""

		new_plgs = [] # Holds the cleaned plugins

		if self.get_setting("Core", "loadplugins", TRUE) and _load_plugins:
			plgs = plugin._loaded_plugins

			for p in plgs:
				a = p.copy()
				del a['function']

				# If it's a sorting or filter plugin don't pass the plugin value
				# The interface should call 'set_ui_sorting_filter' instead
				if (a['type'] == "task_sorting") or (a['type'] == "task_filter"):
					del a['value']

				new_plgs.append(a)

		return new_plgs

	def __dup_task_list(self):
		"""Create a duplicate of the task list"""

		list = self.__tasks
		new_list = []

		for task in list:
			new_list.append(task.copy())

		return new_list

	def interface_call(self, ui, command, *args):
		"""Make a call to an interface that is bound. If no interface is
		specified, then the command is sent to all."""

		if ui == ALL:
			for interface in self.__interfaces:
				interface.call(command, args)

		elif ui in self.__interfaces:
			# Only selected ui's should send a return
			return ui.call(command, args)

	def get_about_text(self, ui=None):
		"""Return the text that should be displayed in the 'About' box"""

		# This makes the text manageable. Even if the formatting looks screwed up.
		return """Authors:\tBrian Bernas (click81@softhome.net)
\tMiles Wiehahn (bigben_miles@hotmail.com)

This software is distributed under the standard Python license.
"""

	# Built-in sorting and filter functions
	def set_ui_sorting_filter(self, ui, sorting=None, filter=None, group=None):
		# Get the sorting and filter plugins
		sort_plgs = []
		filter_plgs = []
		for p in plugin._loaded_plugins:
			if p['type'] == "task_sorting": sort_plgs.append(p)
			elif p['type'] == "task_filter": filter_plgs.append(p)

		# Create a duplicate task list to work with
		tasks = self.__dup_task_list()

		# If a group was specified get all tasks in that group
		if group and (group != 'ALL'):
			group = unicode_to_utf8(group)
			new_tasks = []

			for t in tasks:
				if group in t["Group"]: new_tasks.append(t)

			tasks = new_tasks
			del new_tasks

		# Run the filters first so there will be less to sort
		# Filters
		if filter == "show_unfinished":
			new_tasks = []
			for t in tasks:
				if not t["Finished"]: new_tasks.append(t)

			tasks = new_tasks
			del new_tasks

		else:
			for p in filter_plgs:
				if filter == p['name']:
					tasks = p['function'](tasks)
					break

		# Sorting
		if sorting == "alphabetical":
			# Get the task names
			n = []
			for t in tasks:
				n.append(t["Name"])
			n.sort() # sort the names

			# Re build the list based on the newly sorted names
			new_tasks = []
			for a in n:
				for t in tasks:
					if t["Name"] == a: new_tasks.append(t)

			del tasks
			tasks = new_tasks

		elif sorting == "priority":
			new_tasks = []
			for i in range(5, -1, -1):
				for t in tasks:
					if i == 0:
						if t["Finished"]:
							new_tasks.append(t)
					elif t["Priority"] == i:
						if not t["Finished"]:
							new_tasks.append(t)
			del tasks
			tasks = new_tasks

		else:
			for p in sort_plgs:
				if sorting == p['name']:
					tasks = p['function'](tasks)
					break

		# Make sure that the plugin return what it was supposed to
		if not type(tasks) == type([]):
			self.interface_call(ui, "display_error", _("Plugin Error"),
				_("The plugin: '%s' did not execute correctly.\n"\
				"It is recommened that you contact the creator of this plugin and make sure that"\
				" they are aware of this problem." %p['name']))
			return

		# Get the names of the available tasks
		names = []
		for t in tasks:
			names.append(t["Name"])

		# Tell the interface what tasks are available
		self.interface_call(ui, "update_list_items", names)

	def __get_task_groups(self):
		# Create a list of all task groups
		groups = []

		for t in self.__tasks:
			if t["Group"]:
				for g in t["Group"]:
					if not g in groups:
						groups.append(g)

		# Sort the group list alphabetically
		groups.sort()

		return groups

	def fmttime(self, t):
		"""Format a time structure based on the user's settings"""

		fmt = self.get_setting("Core", "timeformat", "%Y-%m-%d %H:%I:%S")

		t = check_time_values(t)

		if t:
			# If there isn't any time info, then don't include it in the format
			fs = fmt.split()
			if len(fs) > 1:
				if t[3] == 0 and t[4] == 0 and t[5] == 0:
					fmt = fs[0]

			return time.strftime(fmt, t)

		return ''

	def __quit(self):
		"""Stop all proccesses, save the configuration file, and destroy all
		task instances."""

		self.__clear_tmpfile()

		if _load_config:
			self.__save_config()
		del self.__interfaces
		del self.__tasks	# Ah, just for kicks

#-------------------------------------------------------------------------------
# The standard file format
class TaskFile:
	def __init__(self, file):
		self.__file = file

	def read_next_task(self):
		store = FALSE	# Determines when task values should be stored
		attrs = []	# The task attributes

		while 1:
			line = self.__file.readline()
			if not line: break
			line = line.strip()
			if line == '': continue
			elif line == "[TASK]": store = TRUE
			elif line == "[ENDTASK]" and store: break
			elif store:
				# Parse the string value
				try:
					key, value = line.split("=", 1)
				except ValueError: continue

				# Change "special" characters to plain ones
				value = value.replace("\\n", "\n")
				value = value.replace("\\t", "\t")

				if value == "TRUE": value = TRUE
				elif value == "FALSE": value = FALSE

				# A quick hack for backwards compatibility
				if key == "NoFinish":
					attrs.append(("DisableFinish", value))

				attrs.append((key, value))

		# If its the end of the file then return None
		if attrs == []: return None

		# Return the attributes found
		return attrs

	def save_task(self, task):
		file = self.__file

		file.write("[TASK]\n")

		for name, value in task.items():
			# Change "special" characters so they can be written to one line
			# Just hope that the user doesn't use these
			value = str(value).replace("\n", "\\n")
			value = value.replace("\t", "\\t")

			file.write("%s=%s\n" %(name, value))

		file.write("[ENDTASK]\n")

#-------------------------------------------------------------------------------
_help_text = _("""ToDo Manager
usage: python todo-manager.py [options] [filename]
OPTIONS:
 -N\t\t\t:Startup with a new blank file.
 -I\t\t\t:Don't load the configuration file.
 -P\t\t\t:Don't load plugins.
 --help, -h\t\t:Display this text and exit.
 --version, -v\t\t:Display the version number and exit.
""")

# The wheels on the bus go round and round
def parse_args(args):
	# Parsing the command arguments
	global _load_file, _load_config, _load_plugins, _new_file

	for arg in args:
		if arg == "--help":
			print _help_text
			return FALSE
		elif arg == "--version":
			print "%s - %s" %(appname, version)
			return FALSE
		elif arg[0] == "-":
			# Check for command switches
			for a in arg[1:]:
				if a == "h":
					print _help_text
					return FALSE
				elif a == "v":
					print "%s - %s" %(appname, version)
					return FALSE
				elif a == 'N': _new_file = not _new_file
				elif a == 'I': _load_config = not _load_config
				elif a == 'P': _load_plugins = not _load_plugins
		elif os.path.isfile(arg):
			arg = os.path.join(os.getcwd(), arg)
			_load_file = arg
		else:
			print _("%s does not exist.") %arg
			return FALSE

	return TRUE

# The global application instance. Everything calls this to access the core
app_instance = None

def get_app():
	global app_instance

	# If it doesn't exist then create it. It's that simple
	if not app_instance:
		app_instance = App()

	return app_instance
