# objectlistbox.py
# This software is released under the Python 2.0.1 license
#
# Author: Brian Bernas
#
# $Revision: 1.9 $

from Tkinter import *

TEXT = "TEXT"
OBJECT = "OBJECT"

#-------------------------------------------------------------------------------
class ListboxError(Exception):
	pass

#-------------------------------------------------------------------------------
class ObjectListbox(Frame):
	def __init__(self, master, **kw):
		Frame.__init__(self, master, bd=2, relief=SUNKEN)

		# Check the values in kw
		if not kw.has_key("width"): kw["width"] = 100
		if not kw.has_key("height"): kw["height"] = 100
		if not kw.has_key("background") and not kw.has_key("bg"):
			kw["background"] = None
		elif kw.has_key("bg"):
			kw["background"] = kw["bg"]
		if not kw.has_key("foreground") and not kw.has_key("fg"):
			kw["foreground"] = 'black'
		elif kw.has_key("fg"):
			kw["foreground"] = kw["fg"]
		if not kw.has_key("itemcolor"): kw["itemcolor"] = "gray90"
		if not kw.has_key("selectioncolor"): kw["selectioncolor"] = "#6A7F9E"
		if not kw.has_key("font"): kw["font"] = None

		# Variables
		self.__objects = []
		self.__cursel = -1
		self.__columns = []
		self.__font = kw["font"]
		self.__item_color = kw["itemcolor"]
		self.__selection_color = kw["selectioncolor"]
		self.__foreground = kw["foreground"]
		self.__dragable = FALSE
		self.__moved_object = FALSE	# So the list isn't updated unless it needs to be
		self.__draw_cache = FALSE	# If this is TRUE then the list won't auto-update

		self.__item_leftclick_command = None
		self.__item_leftdrag_command = None
		self.__item_leftdoubleclick_command = None
		self.__item_middleclick_command = None
		self.__item_rightclick_command = None

		# The frame that holds all on the column headers
		self._columnFrame = Frame(self, bd=0)
		self._columnFrame.grid(row=0, sticky=EW)

		# Create the list canvas
		self._canvas = Canvas(self, bd=0, highlightthickness=0, width=kw["width"],
			height=kw["height"], background=kw["background"])
		scroll = Scrollbar(self, command=self._canvas.yview)
		self._canvas.config(yscrollcommand=scroll.set)
		self._canvas.grid(row=1, sticky=NSEW)
		scroll.grid(row=0, column=1, rowspan=2, sticky=NS)

		# Configure the main grid
		self.grid_columnconfigure(0, weight=1, minsize=215)
		self.grid_rowconfigure(1, weight=1)

		# Bindings
		self.bind('<Configure>', self.__OnConfig)

		self._canvas.bind('<B1-Motion>', self.__OnItemLeftMove)
		self._canvas.bind('<ButtonRelease-1>', self.__OnItemLeftRelease)

	def __OnConfig(self, event=None):
		self._canvas.update()
		self.draw()
		self.scrollrange_set()

	def __OnItemLeftClick(self, index):
		self.__cursel = index

		if self.__item_leftclick_command:
			self.__item_leftclick_command(index)

		self.draw()
		self.__dragable = TRUE

	def __OnItemLeftDoubleClick(self, index):
		self.__cursel = index

		if self.__item_leftdoubleclick_command:
			self.__item_leftdoubleclick_command(index)

		self.draw()

	def __OnItemMiddleClick(self, index):
		if self.__item_middleclick_command:
			self.__item_middleclick_command(index)

		self.draw()

	def __OnItemRightClick(self, index):
		if self.__item_rightclick_command:
			self.__item_rightclick_command(index)

		self.draw()

	def __OnItemLeftMove(self, event):
		if self.__dragable:
			canvas = self._canvas
			y = canvas.canvasy(event.y)
			size = len(self.__objects)
			width = canvas.winfo_width()

			pos = int(y)/16 + (y % 16 >= 8)
			if pos > size: pos = size
			pos = pos * 16

			canvas.delete("tmp")
			canvas.create_polygon(0,pos-4,7,pos,0,pos+4, fill='black', tag="tmp")
			canvas.create_line(0, pos, width, pos, fill='black', width=2,  tag="tmp")
			canvas.create_polygon(width,pos-4,width-7,pos,width,pos+4, fill='black', tag="tmp")

			self.__moved_object = TRUE

	def __OnItemLeftRelease(self, event):
		if self.__dragable and self.__moved_object:
			objects = self.__objects
			canvas = self._canvas
			sel = self.__cursel
			y = canvas.canvasy(event.y)
			size = len(objects)

			pos = int(y)/16 + (y % 16 >= 8)
			if pos > size: pos = size

			obj = objects[sel]
			del objects[sel]

			if pos > sel: pos  = pos - 1
			objects.insert(pos, obj)

			canvas.delete("tmp")
			self.__dragable = FALSE
			self.__moved_object = FALSE
			self.__cursel = pos

			if self.__item_leftdrag_command:
				self.__item_leftdrag_command()

			self.draw()

	def scrollrange_set(self):
		height = self._canvas.winfo_height()
		if len(self.__objects) > (height/16):
			self._canvas.config(scrollregion=(0,0,0,len(self.__objects)*16))
		else:
			self._canvas.config(scrollregion=(0,0,0,height))

	def insert_column(self, index, text, **kw):
		# What keywords were passed and so on
		if not kw.has_key("anchor"): kw["anchor"] = CENTER
		if not kw.has_key("width"): kw["width"] = len(text)
		if not kw.has_key("expand"): kw["expand"] = FALSE
		if not kw.has_key("visible"): kw["visible"] = TRUE

		if index == END: index = len(self.__columns)

		params = [text, kw["anchor"], kw["width"], kw["expand"], kw["visible"]]
		self.__columns.insert(index, params)

	def show_column(self, index, value):
		self.__columns[index][4] = value

	def update_columns(self):
		# Destroy the current columns
		children = self._columnFrame.winfo_children()
		for child in children:
			child.destroy()

		# Draw the current column list
		for col in self.__columns:
			if col[4]:
				Label(self._columnFrame, text=col[0], bd=1, relief=RAISED, anchor=col[1],
					width=col[2]).pack(side=LEFT, fill=X, expand=col[3])

		self.draw()

	def insert_text(self, row, column, text, **kw):
		if not kw.has_key("color"): kw["color"] = self.__foreground
		if not kw.has_key("selectcolor"): kw["selectcolor"] = 'white'
		if not kw.has_key("font"): kw["font"] = self.__font

		if row == END: row = self.count()
		colcount = len(self.__columns) -1

		if column > colcount:
			raise ListboxError, "Column %n does not exist." %column
		
		obj = (TEXT, text, kw["color"], kw["selectcolor"], kw["font"])

		if row >= self.count():
			# Create a blank list and insert obj into it
			l = []
			for i in range(colcount):
				l.append(None)
			l.insert(column, obj)
			self.__objects.insert(row, l)

		else:
			self.__objects[row][column] = obj

		self.draw()

	def insert_object(self, row, column, function):
		if row == END: row = self.count()
		colcount = len(self.__columns) -1

		if column > colcount:
			raise ListboxError, "Column %n does not exist." %column

		obj = (OBJECT, function)

		if row >= self.count():
			l = []
			for i in range(colcount):
				l.append(None)
			l.insert(column, obj)
			self.__objects.insert(row, l)

		else:
			self.__objects[row][column] = obj

		self.draw()

	def get_column_text(self, row, column):
		try:
			if self.__objects[row][column][0] == TEXT:
				return self.__objects[row][column][1]
		except (IndexError, TypeError):
			return None

	def get_column_object(self, row, column):
		try:
			if self.__objects[row][column][0] == OBJECT:
				return self.__objects[row][column][1]
		except (IndexError, TypeError):
			return None

	def get_column_data(self, row, column):
		"""Returns the data found at the specified column without any type checks"""
		try:
			return self.__objects[row][column][1]
		except (IndexError, TypeError):
			return None

	def delete_item(self, index):
		if index == -1: return
		if index == ALL:
			del self.__objects
			self.__objects = []
		else:		
			del self.__objects[index]

		self.draw()

	def curselection(self):
		return self.__cursel

	def selection_set(self, index):
		self.__cursel = index
		self.draw()

	def count(self):
		return len(self.__objects)

	def set_item_leftclick(self, command):
		if callable(command):
			self.__item_leftclick_command = command
		else:
			raise Exception, "command must be a callable function."

	def set_item_doubleclick(self, command):
		if callable(command):
			self.__item_leftdoubleclick_command = command
		else:
			raise Exception, "command must be a callable function."

	def set_item_leftdrag(self, command):
		if callable(command):
			self.__item_leftdrag_command = command
		else:
			raise Exception, "command must be a callable function."

	def set_item_middleclick(self, command):
		if callable(command):
			self.__item_middleclick_command = command
		else:
			raise Exception, "command must be a callable function."

	def set_item_rightclick(self, command):
		if callable(command):
			self.__item_rightclick_command = command
		else:
			raise Exception, "command must be a callable function."

	def start_draw_cache(self):
		"""Keep the list from updating automatically.
		This is good to use when you have a large group of items to add to the list."""
		self.__draw_cache = TRUE

	def end_draw_cache(self):
		"""End the draw cache and update the list."""
		self.__draw_cache = FALSE
		self.draw()

	def draw(self):
		# ...and the fun begins
		if self.__draw_cache: return

		canvas = self._canvas
		canvas.delete(ALL)

		index = 0
		for object in self.__objects:
			tag = "item" + `index`
			xpos = 0
			count = -1
			colcount = 0

			select = (self.__cursel == index)

			# Draw the selection rectangle
			width = self._canvas.winfo_width()
			fill = [self.__item_color, self.__selection_color][select]
			canvas.create_rectangle(0, (index*16)+1, width, (index*16)+16, fill=fill, width=0,
				tag=tag)

			for obj in object:
				count = count + 1
				curcol = self.__columns[count]

				# If the column isn't visible move on
				if not curcol[4]: continue

				column = self._columnFrame.winfo_children()[colcount]
				colwidth = column.winfo_width()

				# Alignment stuff
				if curcol[1] == W:
					x = xpos
				elif curcol[1] == CENTER:
					x = colwidth/2 + xpos
				elif curcol[1] == E:
					x = colwidth + xpos

				# Drawing to the canvas
				if obj:
					if obj[0] == TEXT:
						fill = [obj[2], obj[3]][select]

						# Text cropping
						text = obj[1]

						if text:
							crop = column.winfo_width()/6
							if len(text) > crop:
								text = "%s..." %text[:crop-2]

						canvas.create_text(x, (index*16)+8, text=text, anchor=curcol[1], fill=fill,
							font=obj[4], tag=tag)

					elif obj[0] == OBJECT:
						# The function should do it all. Drawing, create windows, etc.
						# Just give it the parameters
						# function(canvas, row, column, x, y, tag)
						func = obj[1]
						
						if callable(func):
							func(canvas, index, count, x, (index*16), tag)

				xpos = colwidth + xpos
				colcount = colcount + 1

			# Tag bindings
			canvas.tag_bind(tag, '<Button-1>', lambda e, s=self, i=index: s.__OnItemLeftClick(i))
			canvas.tag_bind(tag, '<Double-Button-1>', lambda e, s=self, i=index:
				s.__OnItemLeftDoubleClick(i))
			canvas.tag_bind(tag, '<Button-2>', lambda e, s=self, i=index: s.__OnItemMiddleClick(i))
			canvas.tag_bind(tag, '<Button-3>', lambda e, s=self, i=index: s.__OnItemRightClick(i))

			index = index + 1

		self.scrollrange_set()

	def configure(self, *args, **kw):
		# Handle __setitem__ command passes
		if args and not kw:
			kw = args[0]

		for key, value in kw.items():
			# Check for canvas specific options
			if key in ("width", "height", "background", "bg"):
				self._canvas.configure({key: value})
				del kw[key]	# Remove this key so it doesn't get passed to Frame

			# Check for task color options
			else:
				if key == "foreground" or key == "fg":
					self.__foreground = value
					del kw[key]
				if key == "itemcolor":
					self.__item_color = value
					del kw[key]
				if key == "selectioncolor":
					self.__selection_color = value
					del kw[key]
				if key == "font":
					self.__font = value
					del kw[key]

		# Send the left over options to the frame
		Frame.configure(self, kw)
	
	# Alias config to configure since some people don't like to type a lot
	config = configure

	def __setitem__(self, key, value):
		self.configure({key: value})

	def __getitem__(self, key):
		# Check for setting that apply to the canvas
		if key in ("width", "height", "background", "bg"):
			return self._canvas[key]

		# Custom values
		elif key == "foreground" or key == "fg": return self.__foreground
		elif key == "itemcolor": return self.__item_color
		elif key == "selectioncolor": return self.__selection_color
		elif key == "font": return self.__font

		# If the setting isn't directly supported here, pass it to the Frame
		else:
			return Frame.__getitem__(self, key)
